"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CapturingBrowser = void 0;
const events_1 = require("events");
const path_1 = __importDefault(require("path"));
const storycrawler_1 = require("storycrawler");
const errors_1 = require("./errors");
const screenshot_options_helper_1 = require("../shared/screenshot-options-helper");
/**
 *
 * A worker to capture screenshot images.
 *
 **/
class CapturingBrowser extends storycrawler_1.StoryPreviewBrowser {
    /**
     *
     * @override
     *
     * @param opt - Options for Storycap.
     * @param mode - Indicates this worker runs as managed mode or simple mode.
     * @param idx - Worker id.
     *
     **/
    constructor(connection, opt, mode, idx) {
        super(connection, idx, opt, opt.logger);
        this.connection = connection;
        this.opt = opt;
        this.mode = mode;
        this.currentStoryRetryCount = 0;
        this.processedStories = new Set();
        this.currentVariantKey = { isDefault: true, keys: [] };
        this.touched = false;
        this.emitter = new events_1.EventEmitter();
        this.emitter.on('error', e => {
            throw e;
        });
        this.baseScreenshotOptions = (0, screenshot_options_helper_1.createBaseScreenshotOptions)(opt);
    }
    /**
     *
     * @override
     *
     **/
    async boot() {
        await super.boot();
        await this.expose();
        await this.addStyles();
        this.resourceWatcher = new storycrawler_1.ResourceWatcher(this.page).init();
        return this;
    }
    async addStyles() {
        if (this.opt.disableCssAnimation) {
            await this.page.addStyleTag({ path: path_1.default.resolve(__dirname, '../../assets/disable-animation.css') });
        }
    }
    async expose() {
        const exposed = {
            emitCapture: (opt, clientStoryKey) => this.subscribeScreenshotOptions(opt, clientStoryKey),
            getBaseScreenshotOptions: () => this.baseScreenshotOptions,
            getCurrentVariantKey: () => this.currentVariantKey,
            waitBrowserMetricsStable: () => this.waitBrowserMetricsStable('preEmit'),
        };
        await Promise.all(Object.entries(exposed).map(([k, f]) => this.page.exposeFunction(k, f)));
    }
    async reload() {
        await this.page.reload();
        await (0, storycrawler_1.sleep)(this.opt.viewportDelay);
        await this.addStyles();
    }
    async waitIfTouched() {
        if (!this.touched)
            return;
        await (0, storycrawler_1.sleep)(this.opt.stateChangeDelay);
    }
    async resetIfTouched(screenshotOptions) {
        const story = this.currentStory;
        if (!this.touched || !story)
            return;
        this.debug('Reset story because page state got dirty in this request.', this.currentRequestId);
        await this.setCurrentStory(story, { forceRerender: true });
        // Clear the browser's mouse state.
        if (screenshotOptions.click) {
            // Workaround for an issue where puppeteer fails to find a matching element.
            // https://github.com/reg-viz/storycap/issues/712
            await this.page.evaluate(() => { });
            await this.page.$eval(screenshotOptions.click, (e) => e === null || e === void 0 ? void 0 : e.blur());
        }
        if (screenshotOptions.focus) {
            // Workaround for an issue where puppeteer fails to find a matching element.
            // https://github.com/reg-viz/storycap/issues/712
            await this.page.evaluate(() => { });
            await this.page.$eval(screenshotOptions.focus, (e) => e === null || e === void 0 ? void 0 : e.blur());
        }
        await this.page.mouse.move(0, 0);
        await this.page.mouse.click(0, 0);
        this.touched = false;
        return;
    }
    async subscribeScreenshotOptions(opt, clientStoryKey) {
        if (this.touched)
            return;
        if (!this.currentStory) {
            this.emitter.emit('error', new errors_1.InvalidCurrentStoryStateError());
            return;
        }
        // Sometimes preview window emits options before completion of change story.
        // So asserts story identifiers between this class hold and sent from browser and skips procedure if they're not equal.
        if (this.currentStory.id !== clientStoryKey) {
            this.debug('This options was sent from previous story', this.currentStory.id, clientStoryKey);
            return;
        }
        // Sometimes preview window emits options twice for a story.
        // The second(or more) options should be ignore because we should guarantee just one PNG for each request.
        if (this.processedStories.has(this.currentRequestId)) {
            this.debug('This story was already processed:', this.currentRequestId, this.currentStory.kind, this.currentStory.story, this.currentVariantKey, JSON.stringify(opt));
            return;
        }
        this.processedStories.add(this.currentRequestId);
        this.debug('Start to process to screenshot story:', this.currentRequestId, this.currentStory.kind, this.currentStory.story, this.currentVariantKey, JSON.stringify(opt));
        this.emitter.emit('screenshotOptions', opt);
    }
    async waitForOptionsFromBrowser() {
        return new Promise((resolve, reject) => {
            const id = setTimeout(() => {
                this.emitter.removeAllListeners();
                if (!this.currentStory) {
                    reject(new errors_1.InvalidCurrentStoryStateError());
                    return;
                }
                if (this.currentStoryRetryCount < this.opt.captureMaxRetryCount) {
                    this.opt.logger.warn(`Capture timeout exceeded in ${this.opt.captureTimeout + ''} msec. Retry to screenshot this story after this sequence.`, this.currentStory.kind, this.currentStory.story);
                    resolve(undefined);
                    return;
                }
                reject(new errors_1.ScreenshotTimeoutError(this.opt.captureTimeout, this.currentStory));
            }, this.opt.captureTimeout);
            const cb = (opt) => {
                clearTimeout(id);
                this.emitter.removeAllListeners();
                resolve(opt);
            };
            this.emitter.once('screenshotOptions', cb);
        });
    }
    async setViewport(opt) {
        var _a, _b;
        if (!this.currentStory) {
            throw new errors_1.InvalidCurrentStoryStateError();
        }
        let nextViewport;
        if (typeof opt.viewport === 'string') {
            if (opt.viewport.match(/^\d+$/)) {
                // For case such as `--viewport "800"`.
                nextViewport = { width: +opt.viewport, height: 600 };
            }
            else if (opt.viewport.match(/^\d+x\d+$/)) {
                // For case such as `--viewport "800x600"`.
                const [w, h] = opt.viewport.split('x');
                nextViewport = { width: +w, height: +h };
            }
            else {
                // Handle as Puppeteer device descriptor.
                const hit = (0, storycrawler_1.getDeviceDescriptors)().find(d => d.name === opt.viewport);
                if (!hit) {
                    this.opt.logger.warn(`Skip screenshot for ${this.opt.logger.color.yellow(JSON.stringify(this.currentStory))} because the viewport ${this.opt.logger.color.magenta(opt.viewport)} is not registered in 'puppeteer/DeviceDescriptor'.`);
                    return false;
                }
                nextViewport = hit.viewport;
            }
        }
        else {
            nextViewport = opt.viewport;
        }
        // Sometimes, `page.screenshot` is completed before applying viewport unfortunately.
        // So we compare the current viewport with the next viewport and wait for `opt.viewportDelay` time if they are different.
        if (!this.viewport || JSON.stringify(this.viewport) !== JSON.stringify(nextViewport)) {
            this.debug('Change viewport', JSON.stringify(nextViewport));
            await this.page.setViewport(nextViewport);
            // Setting isMobile or hasTouch properties will reload the page.
            // See also https://github.com/puppeteer/puppeteer/blob/main/docs/api.md#pagesetviewportviewport
            const willBeReloaded = nextViewport.isMobile !== ((_a = this.viewport) === null || _a === void 0 ? void 0 : _a.isMobile) || nextViewport.hasTouch !== ((_b = this.viewport) === null || _b === void 0 ? void 0 : _b.hasTouch);
            this.viewport = nextViewport;
            if (willBeReloaded || this.opt.reloadAfterChangeViewport) {
                this.processedStories.delete(this.currentRequestId);
                await Promise.all([this.reload(), this.waitForOptionsFromBrowser()]);
            }
            else {
                await (0, storycrawler_1.sleep)(this.opt.viewportDelay);
            }
        }
        return true;
    }
    async warnIfTargetElementNotFound(selector) {
        const targetElement = await this.page.$(selector);
        if (this.currentStory && !targetElement) {
            this.logger.warn(`No matched element for "${this.logger.color.yellow(selector)}" in story "${this.currentStory.id}".`);
        }
    }
    async setHover(screenshotOptions) {
        if (!screenshotOptions.hover)
            return;
        await this.warnIfTargetElementNotFound(screenshotOptions.hover);
        await this.page.hover(screenshotOptions.hover);
        this.touched = true;
        return;
    }
    async setFocus(screenshotOptions) {
        if (!screenshotOptions.focus)
            return;
        await this.warnIfTargetElementNotFound(screenshotOptions.focus);
        await this.page.focus(screenshotOptions.focus);
        this.touched = true;
        return;
    }
    async setClick(screenshotOptions) {
        if (!screenshotOptions.click)
            return;
        await this.warnIfTargetElementNotFound(screenshotOptions.click);
        await this.page.click(screenshotOptions.click);
        this.touched = true;
        return;
    }
    async waitForResources(screenshotOptions) {
        if (!screenshotOptions.waitAssets && !screenshotOptions.waitImages)
            return;
        this.debug('Wait for requested resources resolved', this.resourceWatcher.getRequestedUrls());
        await this.resourceWatcher.waitForRequestsComplete();
    }
    async waitBrowserMetricsStable(phase) {
        const mw = new storycrawler_1.MetricsWatcher(this.page, this.opt.metricsWatchRetryCount);
        const checkCountUntillStable = await mw.waitForStable();
        this.debug(`[${phase}] Browser metrics got stable in ${checkCountUntillStable} times checks.`);
        if (checkCountUntillStable >= this.opt.metricsWatchRetryCount) {
            this.opt.logger.warn(`Metrics is not stable while ${this.opt.metricsWatchRetryCount} times. ${this.opt.logger.color.yellow(JSON.stringify(this.currentStory))}`);
        }
    }
    logInvalidVariantKeysReason(reason) {
        if (reason) {
            if (reason.type === 'notFound') {
                this.logger.warn(`Invalid variants. The variant key '${reason.to}' does not exist(story id: ${this.currentStory.id}).`);
            }
            else if (reason.type === 'circular') {
                this.logger.warn(`Invalid variants. Reference ${reason.refs.join(' -> ')} is circular(story id: ${this.currentStory.id}).`);
            }
        }
    }
    /**
     * Captures screenshot as a PNG image buffer from a story.
     *
     * @param requestId - Represents an identifier for the screenshot
     * @param variantKey - Variant identifier for the screenshot
     * @param retryCount - The number which represents how many attempting to capture this story and variant
     *
     * @returns PNG buffer, whether the capturing process is succeeded or not, additional variant keys if they are emitted, and file name suffix for default the default variant.
     *
     * @remarks
     *
     * - Throws an error if `retryCount` is equal to `opt.captureMaxRetryCount` and this capturing process is failed
     *
     **/
    async screenshot(requestId, story, variantKey, retryCount) {
        var _a;
        this.currentRequestId = requestId;
        this.currentVariantKey = variantKey;
        this.currentStoryRetryCount = retryCount;
        let emittedScreenshotOptions;
        this.resourceWatcher.clear();
        await this.setCurrentStory(story, { forceRerender: true });
        if (this.mode === 'managed') {
            // Screenshot options are emitted form the browser process when managed mode.
            emittedScreenshotOptions = await this.waitForOptionsFromBrowser();
            if (!this.currentStory) {
                throw new errors_1.InvalidCurrentStoryStateError();
            }
            if (!emittedScreenshotOptions) {
                // End this capturing process as failure of timeout if emitter don't resolve screenshot options.
                return { buffer: null, succeeded: false, variantKeysToPush: [], defaultVariantSuffix: '' };
            }
        }
        else {
            await (0, storycrawler_1.sleep)(this.opt.delay);
            await this.waitBrowserMetricsStable('preEmit');
            // Use only `baseScreenshotOptions` when simple mode.
            emittedScreenshotOptions = (0, screenshot_options_helper_1.pickupWithVariantKey)(this.baseScreenshotOptions, this.currentVariantKey);
        }
        const mergedScreenshotOptions = (0, screenshot_options_helper_1.mergeScreenshotOptions)(this.baseScreenshotOptions, emittedScreenshotOptions);
        // Get keys for variants included in the screenshot options in order to queue capturing them after this sequence.
        const [invalidReason, keys] = (0, screenshot_options_helper_1.extractVariantKeys)(mergedScreenshotOptions);
        const variantKeysToPush = this.currentVariantKey.isDefault ? keys : [];
        this.logInvalidVariantKeysReason(invalidReason);
        // End this capturing process as success if `skip` set true.
        if (mergedScreenshotOptions.skip) {
            await this.waitForDebugInput();
            return { buffer: null, succeeded: true, variantKeysToPush, defaultVariantSuffix: '' };
        }
        this.touched = false;
        // Change browser's viewport if needed.
        const vpChanged = await this.setViewport(mergedScreenshotOptions);
        // Skip to capture if the viewport option is invalid.
        if (!vpChanged)
            return { buffer: null, succeeded: true, variantKeysToPush: [], defaultVariantSuffix: '' };
        // Modify elements state.
        await this.setHover(mergedScreenshotOptions);
        await this.setFocus(mergedScreenshotOptions);
        await this.setClick(mergedScreenshotOptions);
        await this.waitIfTouched();
        // Wait until browser main thread gets stable.
        await this.waitForResources(mergedScreenshotOptions);
        await this.waitBrowserMetricsStable('postEmit');
        await this.page.evaluate(() => new Promise(res => window.requestIdleCallback(res, { timeout: 3000 })));
        // Get PNG image buffer
        const rawBuffer = await this.page.screenshot({
            fullPage: emittedScreenshotOptions.fullPage,
            omitBackground: emittedScreenshotOptions.omitBackground,
            captureBeyondViewport: emittedScreenshotOptions.captureBeyondViewport,
            clip: (_a = emittedScreenshotOptions.clip) !== null && _a !== void 0 ? _a : undefined,
        });
        let buffer = null;
        if (Buffer.isBuffer(rawBuffer)) {
            buffer = rawBuffer;
        }
        // We should reset elements state(e.g. focusing, hovering, clicking) for future screenshot for this story.
        await this.resetIfTouched(mergedScreenshotOptions);
        await this.waitForDebugInput();
        return {
            buffer,
            succeeded: true,
            variantKeysToPush,
            defaultVariantSuffix: emittedScreenshotOptions.defaultVariantSuffix,
        };
    }
}
exports.CapturingBrowser = CapturingBrowser;
